/*
 *  Gazebo - Outdoor Multi-Robot Simulator
 *  Copyright (C) 2003  
 *     Nate Koenig & Andrew Howard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * Desc: 3D position interface.
 * Author: Sachin Chitta and John Hsu
 * Date: 10 June 2008
 * SVN: $Id$
 */
#ifndef GAZEBO_ROS_GRASP_HACK_HH
#define GAZEBO_ROS_GRASP_HACK_HH

#include <ros/callback_queue.h>
#include <ros/advertise_options.h>

#include <gazebo/Controller.hh>
#include <gazebo/Entity.hh>
#include <gazebo/Model.hh>
#include <gazebo/Body.hh>
#include <gazebo/Joint.hh>
#include <gazebo/Param.hh>
#include <gazebo/Time.hh>
#include <gazebo/Pose3d.hh>

#include <ros/ros.h>
#include "boost/thread/mutex.hpp"
#include <std_msgs/String.h>

#include <map>

namespace gazebo
{
   class GazeboRosGraspHack : public Controller
   {

      class GripperPose
      {
        public:
          GripperPose(){this->body = NULL;};
          Body* body;
          Pose3d relative_pose;
      };

      /// \brief Constructor
      public: GazeboRosGraspHack(Entity *parent );

      /// \brief Destructor
      public: virtual ~GazeboRosGraspHack();

      /// \brief Load the controller
      /// \param node XML config node
      protected: virtual void LoadChild(XMLConfigNode *node);

      /// \brief Init the controller
      protected: virtual void InitChild();

      /// \brief Update the controller
      protected: virtual void UpdateChild();

      /// \brief Finalize the controller
      protected: virtual void FiniChild();

      /// \brief The parent Body
      private: Body *myParentBody;

      /// \brief The parent Model
      private: Model *myParentModel;

      /// \brief The gripper links
      private: Body* l_wrist_roll_body;
      private: Body* r_wrist_roll_body;

      /// \brief gripper links and store relative poses for contact correction during grasp
      private: std::map<std::string, GripperPose > gripper_bodies;
      private: std::vector<Pose3d> gripper_relative_poses;

      private: bool l_grasp;
      private: bool r_grasp;

      /// \brief object pose relative to wrist_roll_body during grasp
      private: Pose3d l_grasp_relative_pose;
      private: Pose3d r_grasp_relative_pose;

      /// \brief pointer to ros node
      private: ros::NodeHandle* rosnode_;
      private: ros::Publisher pub_;

      /// \brief topic name
      private: ParamT<std::string> *topicNameP;
      private: std::string topicName;

      private: ParamT<double> *graspCheckRateP;
      private: double graspCheckRate;
      private: Time last_grasp_check_time;

      /// \brief A mutex to lock access to fields that are used in message callbacks
      private: boost::mutex lock;
      private: bool found_gripper_links;
      private: void GetGripperLinks();
      private: void CheckForGrasps();

      /// \brief save last_time
      private: Time last_time;

      /// \brief for setting ROS name space
      private: ParamT<std::string> *robotNamespaceP;
      private: std::string robotNamespace;

      /// \brief Keep track of number of connctions
      private: int graspHackConnectCount;
      private: void GraspHackConnect();
      private: void GraspHackDisconnect();

      private: ros::CallbackQueue grasp_hack_queue_;
      private: void GraspHackQueueThread();
      private: boost::thread callback_queue_thread_;
      private: boost::thread get_gripper_links_thread_;

   };

/** \} */
/// @}


}

#endif

